/*
 *  FireWire_IOMac.h
 *  IFFW-DM
 *
 *  Created by Mathew Chasan on Thu Aug 12 2004.
 *  Copyright (c) 2004 __MyCompanyName__. All rights reserved.
 *
 */


#ifndef __FIREWIRE_IO_MAC_INCLUDE_H
#define __FIREWIRE_IO_MAC_INCLUDE_H
#include <iostream>
#include <IOKit/IOKitLib.h>
#include <IOKit/IOCFPlugin.h>
#include <IOKit/avc/IOFireWireAVCLib.h>
#include <IOKit/avc/IOFireWireAVCConsts.h>
#include <IOKit/firewire/IOFireWireLib.h>
#include <IOKit/firewire/IOFireWireFamilyCommon.h>
#include <CoreFoundation/CoreFoundation.h>
#import <Security/Authorization.h>
#import <Security/AuthorizationTags.h>
#include <unistd.h>
#include <Carbon/Carbon.h>

#define	kCareAboutBusResets false
#define	kMaxBusResetRetries 10
#define delay_tick_ms 1000
#define delay_tick .001

#include "../Source/FireWire_IO.h"
#include <Carbon/Carbon.h>

class Mac1394Device
{
public:
	Mac1394Device();
	virtual ~Mac1394Device();
	
	AuthorizationFlags	myFlags; 
	AuthorizationRef	myAuthorizationRef; 
	AuthorizationItem	myItems; 
	AuthorizationRights myRights;
	
	bool		Open();
	bool		Close();
	bool		AuthorizeUser();
	bool		IsOpened();
	int			Count1394Devices(void);
	
	IOReturn	Write(FWAddress addr, void *temp, UInt32 length, bool careAboutBusResets);
	IOReturn	Read(FWAddress addr, void *temp, UInt32 length, bool careAboutBusResets);

private: 
	bool UnloadFirewireLibrary();
	bool ReloadFirewireLibrary();

	io_object_t				m_TargetService;		
	IOFireWireLibDeviceRef	m_InterfaceForTarget;
	IOCFPlugInInterface**	m_PlugInInterface;
	UInt32					m_Generation;
	
	char					m_ProductName[256];
	char					m_VendorName[256];
	unsigned				m_FW_GUID[2];
	bool					m_AppleFWAudioLoaded;
};

class SectorInfo
{
public:
	SectorInfo()
	{
		m_Offset	= 0;
		m_Len		= 0;
		m_CRC		= 0;
		m_Base		= 0;
	}
	SectorInfo(SectorInfo &src)
	{
		m_Offset	= src.m_Offset;
		m_Len		= src.m_Len;
		m_CRC		= src.m_CRC;
		m_Base		= src.m_Base;
	}
	unsigned int m_Offset,m_Len,m_CRC,m_Base;	
};

class BCDInfo
{
public:
	BCDInfo()
	{
		date[0]					= 0;
		time[0]					= 0;
		model_vendor_text[0]	= 0;
		sw_version_on_board		= 0;
		m_ProtocolVersion		= 0;
		m_BootLoaderVersion		= 0;
		hw_model				= 0;
		hw_version				= 0;
		GUID					= 0;
		BOOTLOADER_DATE			= 0;
		BOOTLOADER_TIME			= 0;
		Debugger_DATE			= 0;
		sw_DATE					= 0;
		sw_TIME					= 0;
		BridgeCoID				= 0;
		Debugger_TIME			= 0;
		hw_modelBCD				= 0;
		node_vendor_idBCD		= 0;
		vendor_id				= 0;
		sw_id					= 0;
		sw_version				= 0;
	}

	char	date[24];
	char	time[24];
	char	model_vendor_text[44];
	int		sw_version_on_board;
	
	unsigned int m_ProtocolVersion, m_BootLoaderVersion, hw_model, hw_version;
	unsigned long long GUID,BOOTLOADER_DATE,BOOTLOADER_TIME,Debugger_DATE,Debugger_TIME;
	unsigned int hw_modelBCD, node_vendor_idBCD, vendor_id, sw_id,sw_version;
	unsigned long long sw_DATE,sw_TIME;
	unsigned long long BridgeCoID;
};


class MacFireWireIO : public IFireWireIO
{
	public:
		MacFireWireIO();
		virtual ~MacFireWireIO();
		
		void		SetBCDImage( unsigned char* data, long size);
		int			CountBridgeCo1394Devices(void) { return m_BCDevice.Count1394Devices();}
		bool		IsCompatibleDevice();
		bool		IsUserAdministrator() {return m_BCDevice.AuthorizeUser();} 
		bool		DownloadBCDFile(); 
		bool		ValidateBCDData() { return m_BCDSoftwareInfoValid;}
		IOReturn	BeginUpdate(unsigned int sector_number);
		IOReturn	EndUpdate();
		IOReturn	GetHardwareBCDInfo(BCDInfo &theInfo);
		IOReturn	RestoreNormalOperation();
		IOReturn	Reset();
		IOReturn	Halt();
		
	private:
		IOReturn	GetOffsetAndLength(unsigned int *offset, unsigned int *length, unsigned int sector_number);
		IOReturn	SendBlock( void* bytes, UInt32 length, UInt32 offset);
	
	protected:
		IOReturn		DecodeBCDHeader(void* bytes, BCDInfo &theInfo);
		bool			m_BCDSoftwareInfoValid;
		unsigned char	*m_BCDData;
		unsigned int	m_BCDSize;
		
		BCDInfo			m_BCDInfo;
		SectorInfo		m_Sectors[3];
		unsigned int	my_Address_Image;
		Mac1394Device	m_BCDevice;
		unsigned short	m_CommID;
		unsigned int	m_SequenceNumber;
};

typedef struct {
	unsigned long protocol;
	unsigned short ID;
	unsigned char command_code;
	unsigned char quadlets;
	unsigned char ops[];
} packet_req;


#endif __FIREWIRE_IO_MAC_INCLUDE_H