#ifndef TYPES_H
#define TYPES_H

#include "XPlatform.h"
#include "XBasicTypes.h"



#include "limits.h"
#ifdef COMMENT
#define CHAR_BIT 8
#define CHAR_MAX 127
#define CHAR_MIN (-128)
#define INT_MAX 2147483647
#define INT_MIN (-2147483647-1)
#define LONG_MAX 2147483647L
#define LONG_MIN (-2147483647L-1L)

#define SCHAR_MAX 127
#define SCHAR_MIN (-128)
#define SHRT_MAX 32767
#define SHRT_MIN (-32768)
#define UCHAR_MAX 255
#define UINT_MAX 4294967295U
#define ULONG_MAX 4294967295UL
#define USHRT_MAX 65535
#define		NULL	0
#endif



//////////////////////////////////////////////////////////////////////////
// Basic Abstract Class
//////////////////////////////////////////////////////////////////////////
class  XUnknown 
{
public:
    XUnknown(void);
    virtual ~XUnknown(void);		// This is the key - virtual destructor for ALL classes
};

typedef XUnknown* pXUnknown;
extern uint32 gObjectCount;				// This is a total count on ALL objects in the system


class  XPoint
{
public:
	XPoint(long theX = 0, long theY = 0) : x(theX), y(theY) { }
	XPoint(const XPoint &pt) : x(pt.x), y(pt.y) { }

	inline XPoint& operator = (const XPoint &thePoint)
	{
		if(this == &thePoint)
			return *this;		// Trying to assign to itself

		x = thePoint.x;
		y = thePoint.y;
		return *this;
	}
	inline void Offset(long dx, long dy)
	{
		x	+= dx;
		y 	+= dy;
	}
	inline void Set(long _x, long _y)
	{ 
		x = _x; y = _y;
	}

	inline XPoint &operator+=(const XPoint &pt)
	{
		x += pt.x;
		y += pt.y;
		return *this;
	}

	inline XPoint &operator-=(const XPoint &pt)
	{
		x -= pt.x;
		y -= pt.y;
		return *this;
	}

	inline bool operator==(const XPoint &thePoint)
	{
		return (thePoint.x == x && thePoint.y == y);
	}
    bool	  operator<( const XPoint& cPoint ) const  { return( y < cPoint.y || ( y == cPoint.y && x < cPoint.x ) ); }
    bool	  operator>( const XPoint& cPoint ) const  { return( y > cPoint.y || ( y == cPoint.y && x > cPoint.x ) ); }
    bool	  operator!=( const XPoint& cPoint ) const { return( y != cPoint.y || x != cPoint.x ); }

	long x;
	long y;
};

// JI: we must use special points, not XPoints, cause they only store longs, which aren't big enough.
class  XDoublePoint
{
public:
	XDoublePoint(double theX = 0.0, double theY = 0.0) : x(theX), y(theY) {}
	XDoublePoint(const XDoublePoint &pt) : x(pt.x), y(pt.y) { }
	XDoublePoint(const XPoint &pt) : x((double)pt.x), y((double)pt.y) {}

	inline XDoublePoint& operator = (const XPoint &thePoint)
	{
		x = (double)thePoint.x;
		y = (double)thePoint.y;
		return *this;
	}

	inline XDoublePoint& operator = (const XDoublePoint &thePoint)
	{
		if(this == &thePoint)
			return *this;		// Trying to assign to itself

		x = thePoint.x;
		y = thePoint.y;
		return *this;
	}

	inline void Offset(double dx, double dy)
	{
		x	+= dx;
		y 	+= dy;
	}
	inline void Set(double _x, double _y)
	{ 
		x = _x; y = _y;
	}

	inline XDoublePoint &operator+=(const XDoublePoint &pt)
	{
		x += pt.x;
		y += pt.y;
		return *this;
	}

	inline XDoublePoint &operator-=(const XDoublePoint &pt)
	{
		x -= pt.x;
		y -= pt.y;
		return *this;
	}

	inline bool operator==(const XDoublePoint &thePoint)
	{
		return (thePoint.x == x && thePoint.y == y);
	}
	bool	  operator<( const XDoublePoint& cPoint ) const  { return( y < cPoint.y || ( y == cPoint.y && x < cPoint.x ) ); }
	bool	  operator>( const XDoublePoint& cPoint ) const  { return( y > cPoint.y || ( y == cPoint.y && x > cPoint.x ) ); }
	bool	  operator!=( const XDoublePoint& cPoint ) const { return( y != cPoint.y || x != cPoint.x ); }

	double	x;
	double	y;
};

// JJ - These operators are SLOW - so from now on we always use += or -= operators and a scratch variable if necessary
//XPoint operator +(const XPoint& pt1, const XPoint& pt2);
//XPoint operator -(const XPoint& pt1, const XPoint& pt2);

#ifdef MAC
#include <Carbon/Carbon.h>
typedef Rect NativeRect;
typedef char * LPSTR;
#endif

#ifdef WIN
typedef RECT NativeRect;
#endif

class  XRect
{
public:
	XRect(void) : left(0), right(0), top(0), bottom(0)	{ }
	XRect(long newLeft, long newTop, long newRight, long newBottom)
		 : left(newLeft), right(newRight), top(newTop), bottom(newBottom)	{ }

	// Copy constructor....
	XRect(const XRect &theRect)
		 : left(theRect.left), right(theRect.right), top(theRect.top), bottom(theRect.bottom)	{ }

	//Copy Constructor for Native structs
	XRect(const NativeRect &theRect)
		 : left(theRect.left), right(theRect.right), top(theRect.top), bottom(theRect.bottom)	{ }

 //   bool	IsValid() const	{ return(left < right && top < bottom);	 }
    inline bool	IsValid() const	{ return( left <= right && top <= bottom );	 }

	inline long	GetHeight(void) const { return xabs(bottom - top);}
	inline long	GetWidth(void) const { return xabs(right - left);}
	inline long	Height(void) const { return GetHeight();}
	inline long	Width(void) const { return GetWidth();}

	inline XPoint	GetLeftTop(void) const { return XPoint(left, top); }
    inline XPoint	GetRightBottom(void) const	{ return(XPoint(right, bottom)); }
	inline XPoint	GetRightTop(void) const { return XPoint(right, top); }
    inline XPoint	GetLeftBottom(void) const	{ return(XPoint(left, bottom)); }

    inline XRect	GetBounds(void) const	{ return( XRect(0, 0, right - left, bottom - top));	}

    inline XRect&	Resize(long nLeft, long nTop, long nRight, long nBottom) 
	{
		left += nLeft; top += nTop; right += nRight; bottom += nBottom;
		return(*this);
    }

	inline XRect& operator =(const XRect &theRect)
	{
		if(this == &theRect)
			return *this;		// Trying to assign to itself

		left    = theRect.left;
		right   = theRect.right;
		top     = theRect.top;
		bottom  = theRect.bottom;
		return *this;
	}
	inline XRect& operator =(const NativeRect &theRect)
	{
		left    = theRect.left;
		right   = theRect.right;
		top     = theRect.top;
		bottom  = theRect.bottom;
		return *this;
	}
	inline  XRect operator&(const XRect& cRect) const
    { 
		return(XRect(xmax(left, cRect.left), xmax(top, cRect.top), xmin(right, cRect.right), xmin(bottom, cRect.bottom))); 
	}

    inline void operator&=(const XRect& cRect)
    { 
		left = xmax(left, cRect.left); 
		top = xmax(top, cRect.top);
		right = xmin(right, cRect.right); 
		bottom = xmin(bottom, cRect.bottom); 
	}

    inline XRect operator|(const XRect& cRect) const
    { 
		return(XRect(xmin(left, cRect.left), xmin(top, cRect.top), xmax(right, cRect.right), xmax(bottom, cRect.bottom))); 
	}

    inline void operator|=(const XRect& cRect)
    { 
		left = xmin(left, cRect.left); 
		top = xmin(top, cRect.top);
		right = xmax(right, cRect.right); 
		bottom = xmax(bottom, cRect.bottom); 
	}

   inline XRect operator|(const XPoint& cPoint) const
    { 
		return(XRect(xmin(left, cPoint.x), xmin(top, cPoint.y), xmax(right, cPoint.x), xmax(bottom, cPoint.y))); 
	}

    inline void operator|=(const XPoint& cPoint)
    { 
		left = xmin(left, cPoint.x); top =  xmin(top, cPoint.y); right = xmax(right, cPoint.x); bottom = xmax(bottom, cPoint.y); 
	}

	inline void ZeroOrigin(void)
	{
		right	-= left;
		left	= 0;
		bottom	-= top;
		top		= 0;
	}

	inline void AlignRect(void)
	{
		long saveVal;
		if(right < left)
		{
			saveVal = left;
			left	= right;
			right	= saveVal;
		}
		if(bottom < top)
		{
			saveVal = top;
			top		= bottom;
			bottom	= saveVal;
		}
	}

	inline bool IsEmpty(void) const
	{ 
		if(bottom <= top || right <= left)		// This equates an invalid(Empty) RECT
			return true;

		return false;
	}	

	inline void Clear(void)
	{
		left    = 0;
		right   = 0;
		top     = 0;
		bottom  = 0;
	}

	inline void Set(long newLeft, long newTop, long newRight, long newBottom)
	{
		left    = newLeft;
		right   = newRight;
		top     = newTop;
		bottom  = newBottom;
	}

	inline bool operator==(const XRect &theRect)
	{
		return (theRect.left == left && theRect.top == top && theRect.right == right && theRect.bottom == bottom);
	}

	inline bool operator!=(const XRect &theRect)
	{
		return (theRect.left != left || theRect.top != top || theRect.right != right || theRect.bottom != bottom);
	}

	inline XRect & operator+=(const XPoint &pt)
	{
		Offset(pt.x, pt.y);
		return *this;
	}

	inline XRect & operator-=(const XPoint &pt)
	{
		Offset(-pt.x, -pt.y);
		return *this;
	}
	inline bool PointInRect(XPoint thePoint)  const
	{
		if(thePoint.x >= left && thePoint.x <= right && thePoint.y >= top && thePoint.y <= bottom)
       		return true;

		return false;
	}

	inline bool CoordInRect(long x, long y) const
	{
		// This is used for clipping purposes and right & bottom
		// are not part of the rect
		if(x >= left && x < right && y >= top && y < bottom)
       		return true;

		return false;
	}

	inline bool RectInRect(const XRect &theRect)
	{
		XRect dummy;
		XRect tmpRect = *this;
		return tmpRect.Intersect(theRect, dummy);
	}

	inline bool ContainedInRect(XRect &testRect)  const
	{
		if(testRect.top < top)
			return false;

		if(testRect.left < left)
			return false;

		if(testRect.right > right)
			return false;

		if(testRect.bottom > bottom)
			return false;

		return true;
	}

	inline void Offset(long dx, long dy)
	{
		left	+= dx;
		right	+= dx;
		top		+= dy;
		bottom 	+= dy;
	}

	inline void Inset(long dx, long dy)
	{
		left    += dx;
		right   -= dx;
		top     += dy;
		bottom  -= dy;
	}

	inline bool Intersect(const XRect &theRect, XRect &resultRect)  const
	{
		if(theRect.left > right || theRect.right < left || theRect.top > bottom || theRect.bottom < top)
			return false;		// They do not intersect

		resultRect.left = xmax(theRect.left, left);
		resultRect.top = xmax(theRect.top, top);
		resultRect.right = xmin(theRect.right, right);
		resultRect.bottom = xmin(theRect.bottom, bottom);
		return true;
	}


	inline bool DoIntersect(const XRect &theRect)  const
	{
		if(theRect.left > right || theRect.right < left || theRect.top > bottom || theRect.bottom < top)
			return false;		// They do not intersect

		return true;
	}

	inline bool Union(const XRect &rect1, const XRect &rect2)
	{
		left	= xmin(rect1.left,	rect2.left);
		right	= xmax(rect1.right,	rect2.right); 
		top		= xmin(rect1.top,	rect2.top); 
		bottom	= xmax(rect1.bottom,	rect2.bottom);
		return true;
}


    long   left, right, top, bottom;
};


 bool Intersect(const XRect &src1, const XRect &src2, XRect &resultRect);

class  XRGB
{
private:
	uchar r;
	uchar g;
	uchar b;
	uchar a;

public:
	XRGB(uchar theRed = 0, uchar theGreen = 0, uchar theBlue = 0, uchar theAlpha = 255) : 
		r(theRed), g(theGreen), b(theBlue), a(theAlpha)
	{
	}

	inline void ScaleColors(xfloat thePercent)
	{
		// give us a value in percent.. ie., 0.0 to 100.0 .. or greater..

		// doesn't make sense to pass in negative..
		if(thePercent < 0.0)
			thePercent *= -1.0;

		xfloat tempR = ((xfloat)r * thePercent) / 100.0;
		xfloat tempG = ((xfloat)g * thePercent) / 100.0;
		xfloat tempB = ((xfloat)b * thePercent) / 100.0;

		if(tempR > 255.0)
			tempR = 255.0;
		if(tempG > 255.0)
			tempG = 255.0;
		if(tempB > 255.0)
			tempB = 255.0;

		r = (uchar)tempR;
		g = (uchar)tempG;
		b = (uchar)tempB;
	}

	inline void DeltaColors(short theVal)
	{
		if(theVal > 0)
		{
			uchar maxVal = r;
			if(g > maxVal)
				maxVal = g;
			if(b > maxVal)
				maxVal = b;
			
			if(theVal + maxVal > 255)
				theVal = 255 - maxVal;
		}
		else
		{
			uchar minVal = r;
			if(g < minVal)
				minVal = g;
			if(b < minVal)
				minVal = b;
			

			if(theVal + minVal < 0)
				theVal = -minVal;
		}
		r += (uchar) theVal;
		g += (uchar) theVal;
		b += (uchar) theVal;
	}
	// Copy constructor...
	XRGB(const XRGB &theColor) : 
		r(theColor.r), g(theColor.g), b(theColor.b), a(theColor.a)
	{
	}
	inline uchar blue(void) const { return b; }
	inline uchar red(void) const { return r; }
	inline uchar green(void) const {	return g;}
	inline uchar alpha(void) const {	return a;}

	XRGB& operator &= (XRGB& r);
	// Can use this overloaded operator instead....
	XRGB& operator=(const XRGB &theColor)
	{
		if(this == &theColor)
			return *this;
	
		r = theColor.r;
		g = theColor.g;
		b = theColor.b;			
		a = theColor.a;			
		return *this;
	}

	bool operator ==(const XRGB &theColor) const
	{
		return (theColor.r == r && theColor.g == g && theColor.b == b && theColor.a == a);
	}

	bool operator !=(const XRGB &theColor) const
	{
		return (theColor.r != r || theColor.g != g || theColor.b != b || theColor.a != a);
	}

	void Set(uchar theRed, uchar theGreen, uchar theBlue, uchar theAlpha = 255)
	{
		r = theRed;
		g = theGreen;
		b = theBlue;
		a = theAlpha;
	}
};

// Stock Colors
extern 	XRGB gRedRGB;
extern 	XRGB gBlueRGB;
extern 	XRGB gBlackRGB;
extern 	XRGB gWhiteRGB;
extern 	XRGB gOffWhiteRGB;
extern 	XRGB gGrayRGB;
extern 	XRGB gLightGrayRGB;
extern 	XRGB gDarkGrayRGB;
extern 	XRGB gYellowRGB;
extern 	XRGB gDefaultColorKey;

typedef enum
{
    X_QUAL_LSHIFT = 0x01,
    X_QUAL_RSHIFT = 0x02,
    X_QUAL_SHIFT  = X_QUAL_LSHIFT | X_QUAL_RSHIFT,

    X_QUAL_LCTRL  = 0x04,
    X_QUAL_RCTRL  = 0x08,
    X_QUAL_CTRL   = X_QUAL_LCTRL | X_QUAL_RCTRL,

    X_QUAL_LALT   = 0x10,
    X_QUAL_RALT   = 0x20,
    X_QUAL_ALT    = X_QUAL_LALT | X_QUAL_RALT,

    X_QUAL_REPEAT = 0x40

} XQualifiers;

class  KeyModifier
{
public:
	KeyModifier(uint32 flags) : myFlags(flags) { }

	bool HasShift(void) const { return ((myFlags & X_QUAL_SHIFT) != 0); }
	bool HasCtrl(void) const { return ((myFlags & X_QUAL_CTRL) != 0); }
	bool HasAlt(void) const { return ((myFlags & X_QUAL_ALT) != 0); }
	bool IsRepeat(void) const { return ((myFlags & X_QUAL_REPEAT) != 0); }
	
	bool HasMenuKey(void) const { return HasCtrl() || HasAlt();}

	void	SetFlags(uint32 theFlags) { myFlags = theFlags;}
	uint32	GetFlags(void) {return myFlags;}
private:
	KeyModifier();
	uint32 myFlags;
};

// JI: these really shouldn't be needed anymore, now that we're using doubles..
 inline double XMulDiv(double a, double b, double c)
{
	return (a * b) / c;
}

 inline double MulDivRound(double a, double b, double c)
{
	// Use proper rounding.
	return floor(((a * b) / c) + 0.5);
}

 inline double NearMulDivRound(double a, double b, double c)
{
	return floor(((a * b) / c) + 0.01);		// round if within 1% of next integer
}

 inline bool IsOdd(short theValue)
{
	return bool((theValue&1) != 0);
}

 inline long AbsoluteValue(long theNumber)
{
	if(theNumber < 0)
		return -theNumber;

	return theNumber;
}




//////////////////////////////////////////////////////////
#endif  // TYPES_H
